/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

// @(#) $Id: TabularServletServer.java 2089 2007-11-28 13:56:13Z s3460 $
package at.spardat.xma.datasource;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Enumeration;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import at.spardat.enterprise.exc.SysException;
import at.spardat.properties.XProperties;
import at.spardat.xma.RuntimeDefaults;
import at.spardat.xma.exception.Codes;
import at.spardat.xma.monitoring.TimeingEvent;
import at.spardat.xma.plugins.PluginManagerServer;
import at.spardat.xma.session.Transform;
import at.spardat.xma.session.XMASessionServer;

/**
 * This servlets task is to serve tabular data sources at the XMA server. It must
 * be registered under the servlet mapping <tt>tabular</tt> in the applications
 * web.xml-file.
 *
 * @author YSD, 17.06.2003 12:25:43
 */
public class TabularServletServer extends HttpServlet {

    /**
     * @see javax.servlet.http.HttpServlet#doGet(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
     */
    protected void doGet (HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        TimeingEvent          ev = new TimeingEvent (getMeasurementPraefix(request) + "tabular");  // measurement for imcmonitor
        try {
            doGet0 (request, response);
            ev.success();
        } finally {
            ev.failure();
        }
    }

    /**
     * Called by <tt>doGet</tt>.
     */
    protected void doGet0 (HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {

        String              measurePraefix = getMeasurementPraefix(request);

        /**
         * Measurement that indicates the time when the servlet returns the complete table to the client
         */
        TimeingEvent          evLoaded = new TimeingEvent (measurePraefix + "tabularLoaded");

        /**
         * Measurement that indicates the execution time when the servlet returns a not-modified.
         * Either <tt>evLoaded</tt> or <tt>evNotModified</tt> is also closed.
         */
        TimeingEvent          evNotModified = new TimeingEvent (measurePraefix + "tabularNotModified");

        /**
         * Iterate over request parameters and construct a TableSpec object
         */
        TableSpec           tableSpec = new TableSpec();
        Enumeration         paramNames = request.getParameterNames();
        while (paramNames.hasMoreElements()) {
            String key = (String) paramNames.nextElement();
            String val = request.getParameter(key);
            tableSpec.setProperty (key, val);
        }
        // ensure that the necessary properties are specified
        String [] required = new String[] { "type", "_loc", "_man", "_env" };
        for (int i=0; i<required.length; i++)
            if (tableSpec.getProperty(required[i]) == null)
                throw new SysException ("tabular servlet, property " + required[i] + " missing in " + tableSpec.toString())
                          .setCode(Codes.DS_MISSING_REQUEST_PARAM);

        /**
         * make sure that there is a valid XMASession
         */
        XMASessionServer session = XMASessionServer.getXMASession(request.getSession());


        if (session == null)
            throw new SysException ("Client requests a datasource "+tableSpec.toString()+" via tabular servlet but no session exists yet.")
                      .setCode(Codes.DS_INVALID_SESSION);

        /**
         * get server side data source plugin implementation
         */
        ITabularDataSource          dsPlugin = (ITabularDataSource) PluginManagerServer.getInstance().getPlugin(ITabularDataSource.class);

        /**
         * check if the client does a conditional get
         */
        long                        lastModified = ITabularDataSource.UNKNOWN_TIMESTAMP;
        lastModified = request.getDateHeader("If-Modified-Since");
        if (lastModified == -1) lastModified = ITabularDataSource.UNKNOWN_TIMESTAMP;

        /**
         * retrieve the table
         */
        if (!(dsPlugin instanceof XMATabularDataSourceServer)) {
            throw new SysException (dsPlugin.getClass().getName() + " must extend TabularDataSourceServer.")
                      .setCode(Codes.DS_SERVER_WRONG_PLUNGIN_CLASS);
        }

        XMATabularDataSourceServer     dsPluginImpl = (XMATabularDataSourceServer)dsPlugin;
        ProviderResultServer        result = dsPluginImpl.getTableFromServerCache(session, tableSpec, lastModified);
        /**
         * Headers which have to be set regardless of the result
         */
        response.setDateHeader("Date", new java.util.Date().getTime());
        int expireDurationClientSecs = dsPluginImpl.getExpireDurationClientSecs(tableSpec.getType());
        expireDurationClientSecs = Math.max(expireDurationClientSecs, 60);
        response.setDateHeader("Expires", new java.util.Date().getTime() + (long)(expireDurationClientSecs)*1000L);
        // todo include header no-cache and no-store

        if (result == null) {
            /**
             * the last modified time stamp did not change
             */
            response.setStatus (HttpServletResponse.SC_NOT_MODIFIED);
            response.setDateHeader("Last-Modified", lastModified);
            evNotModified.success(); // terminate measurement
        } else {
            /**
             * write the table
             */
            if (result.lastModified_ != XMATabularDataSourceServer.UNKNOWN_TIMESTAMP)
                response.setDateHeader("Last-Modified", result.lastModified_);
            response.setContentType("application/octet-stream");

            ByteArrayOutputStream buf = new ByteArrayOutputStream();
            result.table_.write(buf);
            byte[]          res = buf.toByteArray();  // preliminary result

            boolean         doCompress = doCompress (res.length);
            boolean         doHash = integrityProtectByHash();

            if (doCompress || doHash) {
                // is there some transformation to be applied?
                Transform       trans = new Transform();
                if (doCompress) res = trans.compress(res);
                if (doHash)     res = trans.generateHash(res);
                response.setHeader (Transform.Header, trans.getTransformations());
            }

            response.setHeader("Cache-Control", "no-cache");//RFC 2616 HTTP 1.1
            OutputStream out = response.getOutputStream();
            out.write(res);

            evLoaded.success(); // terminate measurement
        }
    }

    /**
     * Returns if the returned table should be integrity protected by appending
     * a hash to the table.
     */
    private static boolean integrityProtectByHash () {
        XProperties         node = XProperties.getNodeOfPackage("at.spardat.xma.datasource");
        boolean             protect = node.getBoolean("hashIntegrityProtection", RuntimeDefaults.hashIntegrityProtection);
        return protect;
    }

    /**
     * Returns true if a table whose byte stream is of size <tt>size</tt> should be
     * gzip compressed.
     */
    private static boolean doCompress (int size) {
        XProperties         node = XProperties.getNodeOfPackage("at.spardat.xma.datasource");
        int                 threshold = node.getInt("compressionThreshold", RuntimeDefaults.compressionThreshold);
        if (threshold != -1 && size > threshold) return true;
        return false;
    }

    /**
     * Returns the praefix of a measurement variable name that contains
     * the context path of the web application.
     */
    private static String getMeasurementPraefix (HttpServletRequest req) {
        String      contextPath = req.getContextPath();
        if (contextPath.length() == 0) return "app<xma>:";
        if (contextPath.charAt(0) == '/') contextPath = contextPath.substring(1);
        return "app<" + contextPath + ">:";
    }

    // http://localhost:8080/demo/tabular?type=rsc&_loc=de_AT&_man=188&_env=p&bundle=at.spardat.xma.datasource.test
}
