/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * Created on 04.11.2003
 *
 *
 *
 */
package at.spardat.xma.mdl.list;

import java.util.ArrayList;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.TypedEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;

import at.spardat.enterprise.fmt.AParseException;
import at.spardat.enterprise.fmt.IFmt;
import at.spardat.xma.component.ComponentClient;
import at.spardat.xma.mdl.Atom;
import at.spardat.xma.mdl.AttachmentExceptionClient;
import at.spardat.xma.mdl.ModelChangeEvent;
import at.spardat.xma.mdl.UIDelegateClient;
import at.spardat.xma.mdl.ValidationErrorClient;
import at.spardat.xma.mdl.WModel;
import at.spardat.xma.page.EventAdapter;
import at.spardat.xma.page.PageClient;

/**
 * This class is the UIDelegate for a {@link ListWMClient} widget model. It
 * manages the interaction between an SWT-combo or an SWT-list and the widget model.
 * If a Combo is to be attached, the model must not have the style {@link ListWM#S_MULTI_SELECT}.
 * If a List is to be attached, the model must not have the style {@link ListWM#S_NOT_STRICT}.
 *
 * @author s2877
 */
public class ListUIDelegateClient extends UIDelegateClient {

    /**
     * The attached SWT-control. A List or a Combo
     */
    private Control                     control_;

    /**
     * The opional label usually positioned at the left hand side of the list or combo.
     */
    private Label                       label_;

    /**
     * The widget model.
     */
    private ListWMClient             wModel_;

    /**
     * Keep track of whether we are updating the UI.
     */
    private boolean                     updatingUI_ = false;

    /**
     * The editable-property
     */
    private boolean                     editable_ = true;

    /**
     * The enabled-property
     */
    private boolean                     enabled_ = true;

    /** Do prefixsearch or not */
    private boolean                     doPrefixSearch_ = true;

    /** Do a case sensitive or case insensitive prefix search */
    private boolean                     prefixSearchCaseSensitive_ = true;

    private String                      oldPrefix_ = "";

    /**
     * Is non null if the first entry in the control is a key that is
     * not in the domain of values itself. This may happen, if the
     * selection in the model is out of range. This variable
     * then holds the key of this out of domain selection.
     */
    private String                      outOfDomainKey_;

    /**
     * Stores the drop down lists visibleItemCount.
     * If a combo model is set to setEditable(false) then the widgets.setVisibleItemCount() is set to 0
     * and the old value is stored in this property for resetting the widget's value at setEditable(true).
     */
    private int visibleItemCount_;

    /**
     * Constructor
     * @param wModel the widget model to use this UIDelegate with.
     */
    public ListUIDelegateClient (ListWMClient wModel) {
        wModel_ = wModel;
    }

    /**
     * Adds/removes the outOfDomainKey to the list of the combo
     */
    private void outOfDomain2Combo() {
        Combo combo = (Combo)control_;
        if(outOfDomainKey_!=null && combo.getItemCount()>0) {
            combo.remove(0);
        }
        String selection = wModel_.getSelected();
        if(selection!=null && !wModel_.table_.containsKey(selection)) {
            outOfDomainKey_=selection;
            IFmt fmt = wModel_.getFmt();
            if(fmt!=null) {
                combo.add(fmt.format(selection),0);
            } else {
                combo.add(selection,0);
            }
        } else {
            outOfDomainKey_=null;
        }
    }

    /**
     * Adds/removes the outOfDomainKey to the list
     * If the model is strict it does nothing.
     *
     */
    private void outOfDomain2List() {
        if(!wModel_.isStrict()) {
            List list = (List)control_;
            if(outOfDomainKey_!=null && list.getItemCount()>0) {
                list.remove(0);
            }
            String selection = wModel_.getSelected();
            if(selection!=null && !wModel_.table_.containsKey(selection)) {
                outOfDomainKey_=selection;
                IFmt fmt = wModel_.getFmt();
                if(fmt!=null) {
                    list.add(fmt.format(selection),0);
                } else {
                    list.add(selection,0);
                }
            } else {
                outOfDomainKey_=null;
            }
        }
    }

    /**
     * Transfers the entries of the model to the attached widget.
     * It fills the list of coices of the Combo or List.
     */
    private void list2UI() {
        if(control_ instanceof Combo) {
            Combo combo = (Combo) control_;
            combo.removeAll();
            outOfDomain2Combo();
            for(int i=0;i<wModel_.size();i++) {
                Atom value = wModel_.getEntry(i);
                combo.add(value.toString(wModel_.getFmt()));
            }
        } else if(control_ instanceof List) {
            List list = (List) control_;
            list.removeAll();
            outOfDomain2List();
            for(int i=0;i<wModel_.size();i++) {
                Atom value = wModel_.getEntry(i);
                list.add(value.toString(wModel_.getFmt()));
            }
        }
    }

    /**
     * Transfers the selection to the attached widget.
     * It selects the selected entries of the model in
     * the widget. Of course for combos only on entry can
     * be selected.
     */
    protected void selection2UI() {
        if(control_ instanceof Combo) {
            Combo combo = (Combo) control_;
            combo.deselectAll();
            if(wModel_.getSelectionCount()>0) {
                String selected = wModel_.getSelected();
                if(selected != null && wModel_.isUserStrict() && wModel_.table_.containsKey(selected)){
                	Atom value = wModel_.table_.get(selected, 0); //get the value to the key, needed if key value pairs were added //KWF
                	selected = value.toString();
                }
                IFmt fmt = wModel_.getFmt();
                if(fmt!=null) {
                    combo.setText(fmt.format(selected));
                } else {
                    combo.setText(selected);
                }
            }
        } else if(control_ instanceof List) {
            List list = (List) control_;
            list.deselectAll();
            if(wModel_.getSelectionCount()>0) {
                if(outOfDomainKey_!=null) {
                    list.select(0);
                } else {
                    String[] selection = wModel_.getSelection();
                    for(int i=0;i<selection.length;i++) {
                    	list.select(wModel_.indexOf(selection[i])); //indexOf returns the right index for key or value only
                    }
                }
            }
        }
    }


    /**
     * Reads the text of the attached combo and
     * puts it into the selection of the model.
     * This method is only called by prefixSearch(),
     * if the entered text does not match any entry
     * and the model is not strict.
     */
    private void combo2selection() {
        Combo combo = (Combo) control_;
        ArrayList sel = new ArrayList(1);
        try {
            String key = combo.getText();
            IFmt fmt = wModel_.getFmt();
            if(fmt!=null) {
                // convert formated string back to internal representation
                if(!fmt.isOneWay()) {
                    key = fmt.parse(key);
                } else {
                    for(int i=0;i<wModel_.size();i++) {
                        Atom value = wModel_.getEntry(i);
                        String formatted = value.toString(wModel_.getFmt());
                        if(key.equals(formatted)) {
                            key=value.toString();
                            break;
                        }
                    }
                }
            }

            String realkey = wModel_.getKey(key); //if key-value pairs are used in the model, the key might have another value the at the UI, so we get it here
            if(realkey != null && wModel_.isUserStrict()){
            	sel.add(realkey);
            } else{
            	sel.add(key);
            }
        } catch (AParseException ex) {
        }
        wModel_.handle(wModel_.new NewSelectionEvent(sel, true));
    }

    /**
     * Reads the selection of the attached List
     * into the selection of the model.
     */
    private void list2selection() {
        List list = (List) control_;
        int[] indices = list.getSelectionIndices();
        ArrayList sel = new ArrayList(indices.length);
        int offset = outOfDomainKey_==null?0:1;
        for(int i=0;i<indices.length;i++) {
            if(indices[i]==0&&outOfDomainKey_!=null) {
                sel.add(outOfDomainKey_);
            } else {
                //Atom entry = wModel_.getEntry(indices[i]-offset);
                String key = wModel_.getKey(indices[i]-offset);
                sel.add(key);
            }
        }
        wModel_.handle(wModel_.new NewSelectionEvent(sel, true));
    }

    /**
     * Reads the text of the attached combo and
     * puts its corresponding entry into the selection
     * of the model.
     */
    private void prefixSearch() {
        Combo combo = (Combo) control_;
        String prefix = combo.getText();
        boolean found = false;
        if(prefix.length()==0) {
            wModel_.handle(wModel_.new NewSelectionEvent(new ArrayList(1),true));
        } else if(startsWith(oldPrefix_,prefix)) {
            if(!wModel_.isUserStrict()) {
                // keep model synchronized
                combo2selection();
            }
        } else {
            String[] items = combo.getItems();
            // first try exact match
            for(int i=0;i<items.length;i++) {
                if(prefixSearchCaseSensitive_) {
                    if(items[i].equals(prefix)) {
                        found=true;
                        combo2selection();
                        break;
                    }
                } else {
                    if(items[i].equalsIgnoreCase(prefix)) {
                        found=true;
                        combo2selection();
                        break;
                    }
                }
            }
            // then try a prefix match
            for(int i=0;!found&&i<items.length;i++) {
                if(startsWith(items[i],prefix)) {
                    found = true;
                    combo.setText(items[i]);
                    combo.setSelection(new Point(prefix.length(),items[i].length()));
                    int index = outOfDomainKey_==null?i:i-1;
                    if(index<0) {
                        combo2selection();
                    } else {
                        ArrayList sel = new ArrayList(1);
                        String key = wModel_.getKey(index);
                        sel.add(key);
                        wModel_.handle(wModel_.new NewSelectionEvent(sel, true));
                    }
                    break;
                }
            }
            // no match found
            if(!found) {
                if(wModel_.isUserStrict()) {
                    selection2UI();
                    String selected = combo.getText();
                    if(selected.length()>0) {
                        combo.setSelection(new Point(oldPrefix_.length(),selected.length()));
                        prefix=oldPrefix_;
                    } else {
                        prefix="";
                    }
                } else {
                    combo2selection();
                }
            }
        }
        oldPrefix_=prefix;
    }

    /**
     * return true if string starts with prefix
     * prefixSearchCaseSensitive_ determines if this check is case sensitive or not.
     */
    private boolean startsWith(String string, String prefix) {
        if(prefixSearchCaseSensitive_) {
            return string.startsWith(prefix);
        } else {
            if(prefix.length()>string.length()) return false;
            for (int i=0; i<prefix.length(); i++) {
                if (Character.toUpperCase(string.charAt(i))!=Character.toUpperCase(prefix.charAt(i))) {
                    return false;
                }
            }
            return true;
        }
    }

    /**
     * When a Combo field looses focus, its text is adjusted.
     */
    private void comboFocusLost() {
        if ((control_.getStyle() & SWT.READ_ONLY) == 0 && editable_) {
            if(wModel_.isUserStrict()) {
                //assure a valid value
                oldPrefix_="";
                prefixSearch();
            }
            IFmt          fmt = wModel_.getFmt();
            if (fmt != null && !fmt.isOneWay()) {
                // in order to get a normalized representation, we first
                // convert the text into the internal string encoding of
                // the formatter, followed by a conversion back to the
                // external encoding.
                try {
                    String uiText = ((Combo)control_).getText();
                    String internal = fmt.parse(uiText);
                    ((Combo)control_).setText(fmt.format(internal));
                } catch (AParseException ex) {
                }
            }
        }
        if(((Combo)control_).getSelectionIndex()!=-1) {
            selection2UI();
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#handleUIEvent(java.lang.Object,int)
     */
    public void handleUIEvent(Object event, int type) {
        if (!isUIAttached()) return;
        if (updatingUI_) {
            return;
        }
        // sanity check: the Widget in the event must be control_
        if (event instanceof TypedEvent) {
            if (((TypedEvent)event).widget != control_) {
                System.err.println ("widget in event differs from control in UIDelegate; event ignored ...");
                return;
            }
        }

        updatingUI_ = true;
        try {
            if (!editable_) {
                /**
                 * In the case where the combo is not editable, we nevertheless get events since we cannot
                 * avoid it. But we ignore it and restore the combo to the previous state.
                 */
                selection2UI();
            } else if (event instanceof ModifyEvent) {
                if(doPrefixSearch_) {
                    prefixSearch();
                } else {
                    combo2selection();
                }
            } else if (event instanceof SelectionEvent) {
                if(control_ instanceof List) {
                    list2selection();
                } else {
                    combo2selection();
                    oldPrefix_="";
                }
            } else if (event instanceof FocusEvent) {
                if(control_ instanceof Combo) {
                    comboFocusLost();
                    oldPrefix_="";
                }
            }
        } finally {
            updatingUI_ = false;
            updateErrorState();
        }
    }



    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#handleModelChangeEvent(at.spardat.xma.mdl.ModelChangeEvent)
     */
    public void handleModelChangeEvent(ModelChangeEvent event) {
        // if the UI is not attached, do nothing
        if (!isUIAttached()) return;

        updatingUI_ = true;
        try {
            if (event instanceof ListWM.SelectionChangedEvent) {
                if(control_ instanceof Combo) {
                    outOfDomain2Combo();
                } else if (control_ instanceof List) {
                    outOfDomain2List();
                }
                selection2UI();
                oldPrefix_="";
                updateErrorState();
            } else if (event instanceof ListWM.ListChangedEvent) {
                list2UI();
                selection2UI(); // refresh selection
            } else if (event instanceof ListWM.ListItemAddedEvent) {
                int index = ((ListWM.ListItemAddedEvent)event).getIndex();
                String text = wModel_.getEntry(index).toString();
                if(outOfDomainKey_!=null) { index++; }
                if(control_ instanceof Combo) {
                    ((Combo)control_).add(text,index);
                } else {
                    ((List)control_).add(text,index);
                }
            } else if (event instanceof ListWM.ListItemRemovedEvent) {
                int index = ((ListWM.ListItemRemovedEvent)event).getIndex();
                if(outOfDomainKey_!=null) { index++; }
                if(control_ instanceof Combo) {
                    ((Combo)control_).remove(index);
                } else {
                    ((List)control_).remove(index);
                }
            } else if (event instanceof ListWM.ListItemChangedEvent) {
                int index = ((ListWM.ListItemChangedEvent)event).getIndex();
                String text = wModel_.getEntry(index).toString();
                if(outOfDomainKey_!=null) { index++; }
                if(control_ instanceof Combo) {
                    ((Combo)control_).remove(index);
                    ((Combo)control_).add(text,index);
                } else {
                    ((List)control_).remove(index);
                    ((List)control_).add(text,index);
                }
            } else if (event instanceof ListWMClient.FormatterChangedEvent) {
                list2UI();
                selection2UI();
                updateErrorState();
            }

        } finally {
            updatingUI_ = false;
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#getWModel()
     */
    public WModel getWModel() {
        return wModel_;
    }

    // see at.spardat.xma.mdl.UIDelegateClient.createControl()
    public Object createControl(Object parent) {
        if(!(parent instanceof Composite)) throw new IllegalArgumentException("parent must be a composite");
        Composite parentComp = (Composite)parent;
        if(wModel_.isMultiSelect()) {
            return new List (parentComp,SWT.MULTI|SWT.BORDER|SWT.H_SCROLL|SWT.V_SCROLL);
        } else {
            return new Combo(parentComp, SWT.DROP_DOWN|SWT.BORDER);
        }
    }

    // see at.spardat.xma.mdl.UIDelegateClient.addListeners()
    public void addListeners(Object control, EventAdapter adapter) {
        if(control instanceof Combo) {
            ((Combo)control).addFocusListener(adapter);
            ((Combo)control).addModifyListener(adapter);
            ((Combo)control).addSelectionListener(adapter);
        } else if(control instanceof List) {
            ((List)control).addFocusListener(adapter);
            ((List)control).addSelectionListener(adapter);
        } else throw new IllegalArgumentException("unsupported control type: "+control.getClass().getName());
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#isUIAttached()
     */
    public boolean isUIAttached() {
        return control_ != null;
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#attachUI(java.lang.Object, java.lang.Object)
     */
    public void attachUI(Object control, Object label) throws AttachmentExceptionClient {
        if (isUIAttached()) throw new IllegalStateException();

        updatingUI_ = true;
        try {
            if(control instanceof Combo || control instanceof List) {
                control_ = (Control) control;
            } else throw new IllegalArgumentException("control must be a combo or a list");
            // data to gui
            list2UI();
            selection2UI();
            // save label
            if (label != null && label instanceof Label) {
                label_ = (Label) label;
            }
            // reflect editable state on the UI
            setEditable (editable_);  // just to set the background color
            //if the control is already disabled, set this at enabled_ - otherwise do not change enabled_
            //as in the Gen class the GUI designer disable property is set directly at the widget (only for some models)
            enabled_ = control_.getEnabled()?enabled_:false;
            setEnabled(enabled_); //set the model's enabled state at the control

            //for readonly combo disable our own prefix search, as it always selects the FIRST matching entry - even by mouse click.
            //In this case the search mechanism from the Combo Box itself is used.
            //for userstrict but not readonly make sure prefix search is enabled, so it can enforce the strictness.
            //for all other cases leave it to default or programmers choice.
            if(control_ instanceof Combo) {
                boolean readonly = (control_.getStyle()&SWT.READ_ONLY)!=0;
                if(readonly) {
                    setDoPrefixSearch(false);
                } else if(wModel_.isUserStrict()) {
                    setDoPrefixSearch(true);
                }
            }
        } finally {
            updatingUI_ = false;
            updateErrorState();
        }
    }

    /**
     * Sets the widget to the current value of the model. The only situation where
     * widget and model may differ, is if the widget contains a value not accepted by the
     * validiator. In this case the model contains the last known valid value.
     */
    public void resetWidgetFromModel() {
        updatingUI_ = true;
        try {
            selection2UI();
            oldPrefix_="";
            updateErrorState();
        } finally {
            updatingUI_ = false;
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#getUIControl()
     */
    public Object getUIControl() {
        if (!isUIAttached()) throw new IllegalStateException ();
        return control_;
    }

    /**
     * @see at.spardat.xma.mdl.UIDelegateClient#getUILabel()
     */
    public Object getUILabel() {
        if (!isUIAttached()) throw new IllegalStateException ();
        return label_;
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#detachUI()
     */
    public void detachUI() {
        updatingUI_ = true;
        try {
            // clear a previous error state
            PageClient pageModel = wModel_.getPageModelC();
            pageModel.clearValidationErrorImpl(control_);
            pageModel.clearError(control_);
            pageModel.getDialogPage().clearWarning(control_);
            pageModel.getDialogPage().clearInfo(control_);
            enabled_ = control_.getEnabled();
            control_ = null;
            label_ = null;
            outOfDomainKey_=null;
        } finally {
            updatingUI_ = false;
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#isEditable()
     */
    public boolean isEditable() {
        return editable_;
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#setEditable(boolean)
     */
    public void setEditable(boolean what) {
        editable_ = what;
        if (isUIAttached()) {
            updateErrorState();

            if(control_ instanceof Combo) {
                int count = ((Combo)control_).getVisibleItemCount();
                if(count>0){
                    visibleItemCount_ = count;
                }
                if(!editable_){
                    ((Combo)control_).setVisibleItemCount(0);
                }else {
                    ((Combo)control_).setVisibleItemCount(visibleItemCount_);
                }
            }
        }
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#isEnabled()
     */
    public boolean isEnabled () {
        if (isUIAttached()) {
            //always take the state from the widget (if it was set directly on it) as this is the old behavior
            enabled_ = control_.isEnabled();
        }
        return enabled_;
    }

    /* (non-Javadoc)
     * @see at.spardat.xma.mdl.UIDelegateClient#setEnabled(boolean)
     */
    public void setEnabled(boolean what) {
        enabled_ = what;
        if (isUIAttached()) control_.setEnabled(what);
    }

    /**
     * @return if prefix search will be done case sensitive
     * @since 1.7.3
     */
    public boolean isPrefixSearchCaseSensitive() {
        return prefixSearchCaseSensitive_;
    }

    /**
     * determine if prefix seach has to be done case sensitive or not
     * @param prefixSearchCaseSensitive
     * @since 1.7.3
     */
    public void setPrefixSearchCaseSensitive(boolean prefixSearchCaseSensitive) {
        prefixSearchCaseSensitive_ = prefixSearchCaseSensitive;
    }

    /**
     * @return if prefix search will be done on this list model
     * @since 1.8.1
     */
    public boolean isDoPrefixSearch() {
        return doPrefixSearch_;
    }

    /**
     * determine if prefix search has to be done on this model
     * @since 1.8.1
     */
    public void setDoPrefixSearch(boolean doPrefixSearch) {
        if(isUIAttached()&&control_ instanceof Combo) {
            boolean readonly = (control_.getStyle()&SWT.READ_ONLY)!=0;
            if(!readonly&&wModel_.isUserStrict()&&doPrefixSearch==false) {
                throw new IllegalStateException("userstrict combos which are not readonly must have prefix search activated");
            }
        }
        this.doPrefixSearch_ = doPrefixSearch;
    }

    /**
     * This method must be called after every change in an SWT Text control to update
     * the error state information. What must be done here is to check the mandatory
     * property of the widget model.
     */
    void updateErrorState () {
        if(control_ instanceof Combo) {
            Combo combo_ = (Combo) control_;
            PageClient pageModel = wModel_.getPageModelC();
            if (editable_) {
                boolean ok = true;
                String text = combo_.getText();
                IFmt validator = wModel_.getFmt();
                if(validator!=null&&!validator.isOneWay()) {
                    try {
                        // pass the text through the formatter to check if it is legitimate external
                        // input for the widget model
                        validator.parse(text);
                    } catch (AParseException ex) {
                        // construct a ValidationErrorClient object and register it with the Page
                        pageModel.setValidationErrorImpl(new ValidationErrorClient (wModel_, control_, getLabelText(), ex));
                        ok = false;
                    }
                }
                if(ok) {
                    // clear a previous error state
                    pageModel.clearValidationErrorImpl(combo_);
                }
            } else {
                pageModel.clearValidationErrorImpl(combo_);
            }
        }
    }

    /**
     * If <tt>inError</tt>, this method sets the background of the corresponding Control <tt>c</tt>
     * to the error color defined in <tt>ComponentClient</tt>. If <tt>inError</tt> is <tt>false</tt>,
     * the background is reset to the default color.
     */
    public void setErrorColor (boolean inError) {
        if(control_ instanceof Combo) {
            Combo combo_ = (Combo) control_;
            if (editable_) {
                if(inError) {
                    combo_.setBackground (wModel_.getPageModelC().getComponent().getColor(ComponentClient.errorBackground));
                } else {
                    combo_.setBackground(null);
                }
            } else {
                combo_.setBackground (Display.getCurrent().getSystemColor(SWT.COLOR_WIDGET_BACKGROUND));
            }
        }
    }

    /**
     * Returns the text of the label which may be null, if the label is unknown.
     */
    private String getLabelText () {
        return label_ != null ? label_.getText() : null;
    }

}
